// ======== Klassen-Definition: MATRIX (Ergnzung fuer Gleichungssystem) ================
// MATRIX-Klasse fuer Matrizenoperationen vom Typ double
// Rechteckmatrix mat(i,j,k,l) wird spaltenweise gespeichert !!!
// Indizierung von i,j,k,l beginnt bei 0 !!!
// vgl. U. Breymann, C++, Carl Hanser Verlag 1993, S.186 ff
// ======================================================================================
// Copyright: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath/Darmstadt
// experimentelles Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Maengel der Algorithmen!
// Version:  5/1994
// Compiler: C++
// ================ Include-Files =======================================================
#include "pch.h"
#include "framework.h"
#include "MATRIXclass.h"
// ---------------------
// #include "Matrix.h"
// #include "Matrizen.h"
// 
// ============================================================================
// public -> Oeffentliche, virtuelle Mitgliedsfunktionen (Gleichungssysteme)
// ============================================================================
//
// Symmetrische Koeffizientenmatrix nach Gauss staffeln
//                                     -> det = Kmat.nach_gauss_staffeln(Cmat)
// ============================================================================
double MATRIX::nach_gauss_staffeln(MATRIZEN& C)
{
    int mk;
    double d = -1.0;
    // Dimensionen der Matrix C ermitteln
    int ci, cj, ck, cl;
    C.dimension_ermitteln(ci, cj, ck, cl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = this->zeilen_spalten(ki, kj, kk, kl);
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mk, mk);
    C.dimension_aendern(mk, mk);
    // ------------------------------------------------
    //
    d = S3151(*this, C, mk);
    //
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ki, kj, kk, kl);
    C.dimension_aendern(ci, cj, ck, cl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    C.temporaere_matrix_loeschen();
    // ------------------------------------------------
    return d;
};// -------------------------------------------------------------------------
// 
// Symmetrische Koeffizientenmatrix nach Cholesky staffeln
//                                   -> det = Kmat.nach_cholesky_staffeln(Bmat)
// ============================================================================
double MATRIX::nach_cholesky_staffeln(MATRIZEN& B)
{
    int mk;
    double d = -1.0;
    // Dimensionen der Matrix B ermitteln
    int bi, bj, bk, bl;
    B.dimension_ermitteln(bi, bj, bk, bl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = this->zeilen_spalten(ki, kj, kk, kl);
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mk, mk);
    B.dimension_aendern(mk, mk);
    // ------------------------------------------------
    //
    d = S3152(*this, B, mk);
    //
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ki, kj, kk, kl);
    B.dimension_aendern(bi, bj, bk, bl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    B.temporaere_matrix_loeschen();
    // ------------------------------------------------
    return d;
};// -------------------------------------------------------------------------
//
// Unsymmetrische Koeffizientenmatrix nach Gauss staffeln
//                                 -> det = Kmat.nach_gauss_staffeln(Amat,Cmat)
// ============================================================================
double MATRIX::nach_gauss_staffeln(MATRIZEN& A, MATRIZEN& C)
{
    int mk;
    double d = -1.0;
    // Dimensionen der Matrix A ermitteln
    int ai, aj, ak, al;
    A.dimension_ermitteln(ai, aj, ak, al);
    // Dimensionen der Matrix C ermitteln
    int ci, cj, ck, cl;
    C.dimension_ermitteln(ci, cj, ck, cl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = this->zeilen_spalten(ki, kj, kk, kl);
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mk, mk);
    A.dimension_aendern(mk, mk);
    C.dimension_aendern(mk, mk);
    // ------------------------------------------------
    //
    d = S3153(*this, A, C, mk);
    //
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ki, kj, kk, kl);
    A.dimension_aendern(ai, aj, ak, al);
    C.dimension_aendern(ci, cj, ck, cl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    A.temporaere_matrix_loeschen();
    C.temporaere_matrix_loeschen();
    // ------------------------------------------------
    return d;
};// -------------------------------------------------------------------------
//
// Rechte Seite nach Cholesky oder Gauss staffeln
//                                           -> Rmat.staffeln(Bmat,Ymat,symdef)
// ============================================================================
void MATRIX::staffeln(MATRIZEN& B, MATRIZEN& Y, int symdef)
{
    // Dimensionen der Matrix Y ermitteln
    int yi, yj, yk, yl;
    Y.dimension_ermitteln(yi, yj, yk, yl);
    // Zeilen/Spalten der Matrix B ermitteln
    int bi, bj, bk, bl;
    int mb, zr;
    mb = B.zeilen_spalten(bi, bj, bk, bl);
    // Dimensionen der Matrix R ermitteln
    int ri, rj, rk, rl;
    zr = this->dimension_ermitteln(ri, rj, rk, rl);
    // Spalten der Matrix R ermitteln
    zr = zr / mb;
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mb, zr);
    B.dimension_aendern(mb, mb);
    Y.dimension_aendern(mb, zr);
    // ------------------------------------------------
    //
    S3157(*this, B, Y, mb, zr, symdef);
    //
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ri, rj, rk, rl);
    B.dimension_aendern(bi, bj, bk, bl);
    Y.dimension_aendern(yi, yj, yk, yl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    B.temporaere_matrix_loeschen();
};// -------------------------------------------------------------------------
//
// Loesungen nach Cholesky oder Gauss ermitteln
//                                         -> Xmat.loesungen(Bmat,Ymat,symdef)
// ===========================================================================
//*** MATRIX::loesungen(MATRIZEN& B, MATRIZEN& Y, int symdef)       // Korrektur 01/2023
MATRIZEN& MATRIX::loesungen(MATRIZEN& B, MATRIZEN& Y, int symdef)
{
    int mb, zr;
    // Dimensionen der Matrix X ermitteln
    int xi, xj, xk, xl;
    zr = this->dimension_ermitteln(xi, xj, xk, xl);
    // Dimensionen der Matrix Y ermitteln
    int yi, yj, yk, yl;
    Y.dimension_ermitteln(yi, yj, yk, yl);
    // Zeilen/Spalten der Matrix B ermitteln
    int bi, bj, bk, bl;
    mb = B.zeilen_spalten(bi, bj, bk, bl);
    // Spalten der Matrix X ermitteln
    zr = zr / mb;
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mb, zr);
    B.dimension_aendern(mb, mb);
    Y.dimension_aendern(mb, zr);
    // ------------------------------------------------
    // temporare Ergebnismatrix erzeugen
    MATRIZEN *temp = temporaere_matrix_erzeugen(Y, 1);
    //
    S3160(*temp, B, Y, mb, zr, symdef);
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(xi, xj, xk, xl);
    B.dimension_aendern(bi, bj, bk, bl);
    Y.dimension_aendern(yi, yj, yk, yl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    B.temporaere_matrix_loeschen();
    Y.temporaere_matrix_loeschen();
    //
    return *temp;
};// -------------------------------------------------------------------------
//
// Inverse einer   symmetrischen Matrix nach Cholesky oder Gauss
// Inverse einer unsymmetrischen Matrix nach Gauss (ohne Zeilen-/Spalten-
// Vertauschung !!!) bilden
//                                      -> det = Kmat.invertieren(Bmat,symdef)
// ===========================================================================
// Koeffizientenmatrix
// nach Cholesky (nur positiv definite Matrix) oder Gauss invertieren
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// Z(mk,mk) = Matrix fuer die Inverse
//            (Z darf gleich K sein) 
// mk       = Zeilen/Spalten von K, Z
// symdef   =  1 Cholesky
//          =  0 Gauss (symmetrisch)
//          = -1 Gauss (unsymmetrisch)
// Ergebnis: Determinante der Matrix K
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  07.11.2022
// Compiler: C++ 
// ----------------------------------------------------------------------------
double MATRIX::invertieren(MATRIZEN& Z, int symdef)
{
    int mk;
    // Dimensionen der Matrix Z ermitteln
    int zi, zj, zk, zl;
    Z.dimension_ermitteln(zi, zj, zk, zl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = this->zeilen_spalten(ki, kj, kk, kl);
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mk, mk);
    Z.dimension_aendern(mk, mk);
    // ------------------------------------------------
    double det = 0.0;
    //
    det = S3170(*this, Z, mk, symdef);
    // 
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ki, kj, kk, kl);
    Z.dimension_aendern(zi, zj, zk, zl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    Z.temporaere_matrix_loeschen();
    // ------------------------------------------------
    return det;
};// --------------------------------------------------------------------------
//
// Lagerbedingungen in Koeffizientenmatrix einarbeiten
//                                                   -> Kmat.lagern(Lmat,Bmat)
// ===========================================================================
void MATRIX::lagern(MATRIZEN& L, MATRIZEN& B)
{
    int mk;
    // Dimensionen des Vektors L ermitteln
    int li, lj, lk, ll;
    mk = L.dimension_ermitteln(li, lj, lk, ll);
    // Dimensionen der Matrix B ermitteln
    int bi, bj, bk, bl;
    mk = B.dimension_ermitteln(bi, bj, bk, bl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = this->zeilen_spalten(ki, kj, kk, kl);
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    this->dimension_aendern(mk, mk);
    L.dimension_aendern(mk);
    B.dimension_aendern(mk, mk);
    // ------------------------------------------------
    //
    S6115(*this, L, B, mk);
    //
    // ------------------------------------------------
    // Matrix-Dimensionen zuruecksetzen
    this->dimension_aendern(ki, kj, kk, kl);
    L.dimension_aendern(li, lj, lk, ll);
    B.dimension_aendern(bi, bj, bk, bl);
    // ------------------------------------------------
    // ggfl. temporaere Matrizen loeschen
    this->temporaere_matrix_loeschen();
    L.temporaere_matrix_loeschen();
    B.temporaere_matrix_loeschen();
};//--------------------------------------------------------------------------
// 
// ============================================================================
// private -> Private Mitgliedsfunktionen (Gleichungssysteme)
// ============================================================================
// 
// Symmetrische Koeffizientenmatrix nach Gauss staffeln -> S3151(Kmat,Cmat,mk)
// ===========================================================================
double MATRIX::S3151(MATRIZEN &K,MATRIZEN &C,int mk)
{
// Symmetrische Koeffizientenmatrix
// nach Gauss in die gestaffelte Dreiecksform ueberfuehren
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// C(mk,mk) = Matrix fuer die gestaffelte Form (Obermatrix C)
//            (C darf gleich K sein) 
// mk       = Zeilen/Spalten von K, C
// Ergebnis: Determinante der Matrix K
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983
// Compiler: C++ 
// ----------------------------------------------------------------------------
// Algorithmus:
//
// Skalarprodukte
// A(i,j) = K(i,j)-C(l,i)*A(l,j)
// Hauptdiagonale
// A(i,i) = K(i,j)-A(l,i)*A(l,i)/A(l,l)
// Beim Erreichen der Hauptdiagonale wird gleichzeitig gebildet
// C(l,i) =               A(l,i)/A(l,l)
// ----------------------------------------------------------------------------
// Referenztyp wandeln
MATRIX &K_=this->referenz(K);
MATRIX &C_=this->referenz(C);
// --------------------------
// LOKALE VARIABLE
      int i,j,ja,l,la;
      double r,b,det=1.0;
      double *Ci,*Cj;
// ---------------------------------------------------------------------------
//***cout<<"Koeffizientenmatrix K= ";  K.ausgeben(cout);
// ===========================================================================
// SCHLEIFE UEBER DIE SPALTEN j
// ============================
for(j=0;j<mk;j++)
{
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN IN SPALTE j ABFRAGEN
// =====================================
      for(ja=0;ja<mk;ja++)
      {
       if(K.element(ja,j) != 0.0) goto M40;
       C.element(ja,j)=0.0;
      };
// SPALTE J GESTRICHEN -> Ende Schleife j
continue;
// ===========================================================================
// SCHLEIFE UEBER DIE ZEILEN i VON SPALTE j
// ========================================
M40: if(ja > j) goto M280;
//
     for(i=ja;i<=j;i++)
  {
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN VON SPALTE i ABFRAGEN
// ======================================
      if(i == j) {la=ja; goto M60;};
// zuerst Hauptdiagonalelement abfragen
      if(C.element(i,i) != 0.0)
        for(la=ja;la<=i;la++) if(C.element(la,i) != 0.0) goto M60;
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      C.element(i,j)=0.0;
      continue;
//
M60:  r=K.element(i,j);
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
      if(i == j)
// Hauptdiagonalelement
      {
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//***      for(l=la;l<i;l++)
//***      {
// Hauptdiagonale reziprok gespeichert !!!
//***        b=C.element(l,i)*C.element(l,l);
//***        r=r-b*C.element(l,j);
//***        C.element(l,i)=b;
//***      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
      Ci=C_.adresse(la,i);
      Cj=C_.adresse(la,j);
//
      for(l=la;l<i;l++)
      {
// Hauptdiagonale reziprok gespeichert !!!
        b=(*Ci)*C.element(l,l);
        r=r-b*(*Cj);
        (*Ci)=b;
        Ci++;
        Cj++;
      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      }
else
// oberhalb der Hauptdiagonale
      {
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//***      for(l=la;l<i;l++)
//***      {
//***        r=r-C.element(l,i)*C.element(l,j);
//***      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
        r-=skalarprodukt(C_.adresse(la,i),1,C_.adresse(la,j),1,i-la);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      };
// ---------------------------------------------------------------------------
      C.element(i,j)=r;
  };
// ===========================================================================
// HAUPTDIAGONAL-ELEMENT
// =====================
      if(r == 0.0) goto M9999;
// Determinante bilden
      det=det*r;
// Hauptdiagonale reziprok speichern !!!
      C.element(j,j) = 1.0/r;
// ---------------------------------------------------------------------------
// AUS OPTISCHEN GRUENDEN ELEMENTE UNTERHALB DER HAUPTDIAGONALEN
//                                                             AUF NULL SETZEN
      ja=j+1;
M280: if(&K != &C)
      for(i=ja;i<mk;i++) C.element(i,j)=0.0;
// ---------------------------------------------------------------------------
};
// ===========================================================================
//***cout<<"Gestaffelte Koeffizientenmatrix C= ";  C.ausgeben(cout);
//***cout<<"Determinante= "<<det<<'\n';
//
return det;
// ===========================================================================
// FEHLERMELDUNG
// =============
M9999: cerr<<"***ABBRUCH in S3151 (Sym. Matrix nach Gauss staffeln)***"<<'\n';
       cerr<<"   Diagonalelement => "<<r;
//
// STOP
// ====
char c;
cin>>c;
return r;
};// -------------------------------------------------------------------------
//
// Symmetrische Koeffizientenmatrix nach Cholesky staffeln
//                                                      -> S3152(Kmat,Bmat,mk)
// ===========================================================================
double MATRIX::S3152(MATRIZEN &K,MATRIZEN &B,int mk)
{
// Symmetrische und positiv definite Koeffizientenmatrix
// nach Cholesky in die gestaffelte Dreiecksform ueberfuehren
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// B(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix
//            (B darf gleich K sein) 
// mk       = Zeilen/Spalten von K, B
// Ergebnis: Determinante der Matrix K
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983
// Compiler: C++ 
// ----------------------------------------------------------------------------
// Algorithmus:
//
// Skalarprodukte
// B(i,j) = K(i,j)-B(l,i)*B(l,j)
// Hauptdiagonalelement
// B(j,j) = sqrt(B(j,j))
// Elemente oberhalb der Hauptdiagonalen
// B(i,j) = B(i,j)/B(i,i)
// ----------------------------------------------------------------------------
// Referenztyp wandeln
MATRIX &K_=this->referenz(K);
MATRIX &B_=this->referenz(B);
// --------------------------
// LOKALE VARIABLE
//*** int l;
int i,j,ja,la;
double r,det=1.0;
// ---------------------------------------------------------------------------
//***cout<<"Koeffizientenmatrix K= ";  K.ausgeben(cout);
// ===========================================================================
// SCHLEIFE UEBER DIE SPALTEN j
// ============================
for(j=0;j<mk;j++)
{
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN IN SPALTE j ABFRAGEN
// =====================================
      for(ja=0;ja<mk;ja++)
      {
       if(K.element(ja,j) != 0.0) goto M40;
       B.element(ja,j)=0.0;
      };
// SPALTE J GESTRICHEN -> Ende Schleife j
continue;
// ===========================================================================
// SCHLEIFE UEBER DIE ZEILEN i VON SPALTE j
// ========================================
M40: if(ja > j) goto M280;
//
     for(i=ja;i<=j;i++)
  {
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN VON SPALTE i ABFRAGEN
// ======================================
      if(i == j) {la=ja; goto M60;};
// zuerst Hauptdiagonalelement abfragen
      if(B.element(i,i) != 0.0)
        for(la=ja;la<=i;la++) if(B.element(la,i) != 0.0) goto M60;
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      B.element(i,j)=0.0;
      continue;
//
M60:  r=K.element(i,j);
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//***      for(l=la;l<i;l++)
//***      {
//***        r=r-B.element(l,i)*B.element(l,j);
//***      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
        r-=skalarprodukt(B_.adresse(la,i),1,B_.adresse(la,j),1,i-la);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ---------------------------------------------------------------------------
// Bei Erreichen der Hauptdiagonale Schleife verlassen
      if(i == j) break;
// KEINE HAUPTDIAGONALE
// Hauptdiagonale reziprok gespeichert !!!
      B.element(i,j)=r*B.element(i,i);
  };
// ===========================================================================
// HAUPTDIAGONAL-ELEMENT
// =====================
      if(r <= 0.0) goto M9999;
// Determinante bilden
      det=det*r;
// Wurzel aus Hauptdiagonalelement
// Hauptdiagonale reziprok speichern !!!
      B.element(j,j) = 1.0/sqrt(r);
// ---------------------------------------------------------------------------
// AUS OPTISCHEN GRUENDEN ELEMENTE UNTERHALB DER HAUPTDIAGONALEN
//                                                             AUF NULL SETZEN
      ja=j+1;
M280: if(&K != &B)
      for(i=ja;i<mk;i++) B.element(i,j)=0.0;
// ---------------------------------------------------------------------------
};
// ===========================================================================
//***cout<<"Gestaffelte Koeffizientenmatrix B= ";  B.ausgeben(cout);
//***cout<<"Determinante= "<<det<<'\n';
//
return det;
// ===========================================================================
// FEHLERMELDUNG
// =============
M9999: cerr<<"***ABBRUCH in S3152 (Sym. Matrix nach Cholesky staffeln)***"<<'\n';
       cerr<<"   Wurzel aus Diagonalelement => "<<r;
// STOP
// ====
char c;
cin>>c;
return r;
};// -------------------------------------------------------------------------
//
// Unsymmetrische Koeffizientenmatrix nach Gauss staffeln
//                                                 -> S3153(Kmat,Amat,Cmat,mk)
// ===========================================================================
double MATRIX::S3153(MATRIZEN &K,MATRIZEN &A,MATRIZEN &C,int mk)
{
// Unsymmetrische Koeffizientenmatrix
// nach Gauss in die gestaffelte Dreiecksform ueberfuehren
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// A(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix A
//            (A darf gleich K sein) 
// C(mk,mk) = Matrix fuer die gestaffelte Form der Untermatrix C
//            (C darf gleich A sein)
// mk       = Zeilen/Spalten von K, A, C
// Ergebnis: Determinante der Matrix K
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983
// Compiler: C++ 
// ----------------------------------------------------------------------------
// Algorithmus:
//
// Skalarprodukte
// Obermatrix (incl. Hauptdiagonalelement):
//              A(i,j) =   K(i,j)-C(i,l)*A(l,j)
// Untermatrix: C(j,i) = ( K(j,i)-C(l,j)*A(l,i) )/A(j,j)
// ----------------------------------------------------------------------------
// Referenztyp wandeln
MATRIX &K_=this->referenz(K);
MATRIX &A_=this->referenz(A);
MATRIX &C_=this->referenz(C);
// --------------------------
// LOKALE VARIABLE
//*** int l;
int i,j,ja,la,J;
double r,det=1.0;
// ---------------------------------------------------------------------------
//***cout<<"Koeffizientenmatrix K= ";  K.ausgeben(cout);
// ===========================================================================
// SCHLEIFE UEBER DIE SPALTEN und ZEILEN j
// =======================================
for(j=0;j<mk;j++)
{
// =========== Anfang Spalte j ===============================================
// FUEHRENDE NULLEN IN SPALTE j ABFRAGEN
// =====================================
      for(ja=0;ja<=j;ja++)
      {
       if(K.element(ja,j) != 0.0) goto M40;
       A.element(ja,j)=0.0;
      };
// SPALTE J GESTRICHEN -> Ende Spalte j
goto M280;
// ===========================================================================
// SCHLEIFE UEBER DIE ZEILEN i VON SPALTE j
// ========================================
M40:
     for(i=ja;i<=j;i++)
  {
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN VON ZEILE i ABFRAGEN
// ======================================
      if(i == j) {la=ja; goto M60;};
// zuerst Hauptdiagonalelement abfragen
      if(A.element(i,i) != 0.0)
        for(la=ja;la<=i;la++) if(C.element(i,la) != 0.0) goto M60;
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      A.element(i,j)=0.0;
      continue;
//
M60:  r=K.element(i,j);
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//**      for(int l=la;l<i;l++)
//**       r=r-C.element(i,l)*A.element(l,j);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
         r-=skalarprodukt(C_.adresse(i,la),C_.dimi,A_.adresse(la,j),1,i-la);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      A.element(i,j)=r;
  };
// ===========================================================================
// Hauptdiagonalelement der Obermatrix A
// =====================================
//    if(r <= 0.0) goto M9999;    // das war bei Cholesky wegen Wurzel
      if(r == 0.0) goto M9999;
// Determinante bilden
      det=det*r;
// Hauptdiagonale reziprok speichern !!!
      A.element(j,j) = 1.0/r;
// =========== Ende Spalte j ================================================= 
M280: J=j+1;
// ============================================================================
// AUS OPTISCHEN GRUENDEN ELEMENTE in A und C AUF NULL SETZEN
      if(&A != &C)
      {
        if(&K != &A)
        {
// unterhalb der Hauptdiagonalen
          for(i=J;i<mk;i++) A.element(i,j)=0.0;
        };
//
        if(&K != &C)
        {
// Hauptdiagonale auf 1.0 setzen
          if(A.element(j,j) == 0.0) C.element(j,j)=0.0;
                               else C.element(j,j)=1.0;
// rechts von der Hauptdiagonalen
          for(i=J;i<mk;i++) C.element(j,i)=0.0;
        };
      };
// ===========================================================================
// Letzte Zeile -> Ende Schleife j
      if(J == mk) continue;
// =========== Anfang Zeile J=j+1 ============================================
// FUEHRENDE NULLEN IN ZEILE J ABFRAGEN
// ====================================
      for(ja=0;ja<=J;ja++)
      {
       if(K.element(J,ja) != 0.0) goto M340;
       C.element(J,ja)=0.0;
      };
// ZEILE J GESTRICHEN -> Ende Schleife j
continue;
// ===========================================================================
// SCHLEIFE UEBER DIE ZEILEN i VON SPALTE J
// ========================================
M340:
//
     for(i=ja;i<J;i++)
  {
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN VON SPALTE i ABFRAGEN
// ======================================
// zuerst Hauptdiagonalelement abfragen
      if(A.element(i,i) != 0.0)
        for(la=ja;la<=i;la++) if(A.element(la,i) != 0.0) goto M360;
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      C.element(J,i)=0.0;
      continue;
//
M360: r=K.element(J,i);
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//***      for(l=la;l<i;l++)
//***       r=r-C.element(J,l)*A.element(l,i);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
        r-=skalarprodukt(C_.adresse(J,la),C_.dimi,A_.adresse(la,i),1,i-la);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ---------------------------------------------------------------------------
// Hauptdiagonale reziprok gespeichert !!!
      C.element(J,i)=r*A.element(i,i);
  };
// ===========Ende Zeile J ===================================================
};
// ===========================================================================
//***cout<<"Gestaffelte Koeffizientenmatrix A= ";  A.ausgeben(cout);
//***cout<<"Determinante= "<<det<<'\n';
//***cout<<"Gestaffelte Koeffizientenmatrix C= ";  C.ausgeben(cout);
//
return det;
// ===========================================================================
// FEHLERMELDUNG
// =============
M9999: cerr<<"***ABBRUCH in S3153 (Unsym. Matrix nach Gauss staffeln)***"<<'\n';
       cerr<<"   Diagonalelement => "<<r;
// STOP
// ====
char c;
cin>>c;
return r;
};// -------------------------------------------------------------------------
//
// Rechte Seite nach Cholesky oder Gauss staffeln
//                                     -> S3157(Rmat,Bmat,Ymat,mb,zr,symdef)
// =========================================================================
void MATRIX::S3157(MATRIZEN &R,MATRIZEN &B,MATRIZEN &Y,
                                                     int mb,int zr,int symdef)
{
// Rechte Seite eines symmetrischen Gleichungssystems nach
// Cholesky oder Gauss in die gestaffelte Rechteckform ueberfuehren
// R(mb,zr) = Matrix mit voll abgespeicherten Rechten Seiten
// B(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
//            (Untermatrix C bei Gauss unsymmetrisch)  
// Y(mb,zr) = Matrix fuer gestaffelte Form der Rechten Seiten
//            (Y darf gleich R sein)
// mb       = Zeilenzahl von B,R,Y
// zr       = Spaltenzahl der Rechten Seiten R,Y
// symdef= 1 : nach Cholesky staffeln
//       = 0 : nach Gauss (symmetrisch)   staffeln 
//       =-1 : nach Gauss (unsymmetrisch) staffeln
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983
// // Compiler: C++ 
// ----------------------------------------------------------------------------
// Algorithmus:
//
// Skalarprodukte
// Y(i,k)=R(i,k)-B(l,i)*Y(l,k)
// Cholesky: Y(i,k)=Y(i,k)/B(i,i)
// ----------------------------------------------------------------------------
// Referenztyp wandeln
MATRIX &R_=this->referenz(R);
MATRIX &B_=this->referenz(B);
MATRIX &Y_=this->referenz(Y);
// --------------------------
// LOKALE VARIABLE
//*** int l;
int k,ia,i,la;
double r;
// ---------------------------------------------------------------------------
//***cout<<"Rechte Seiten R= ";  R.ausgeben(cout);
//***cout<<"Gestaffelte Koeffizientenmatrix B= ";  B.ausgeben(cout);
// ===========================================================================
// SCHLEIFE UEBER DIE SPALTEN k
// ============================
for(k=0;k<zr;k++)
{
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN IN SPALTE k ABFRAGEN
// =====================================
      for(ia=0;ia<mb;ia++)
      {
       if(R.element(ia,k) != 0.0) goto M40;
       Y.element(ia,k)=0.0;
      };
// SPALTE k GESTRICHEN -> Ende Schleife k
continue;
// ===========================================================================
// SCHLEIFE UEBER DIE ZEILEN i VON SPALTE k
// ========================================
M40: for(i=ia;i<mb;i++)
  {
// ---------------------------------------------------------------------------
// FUEHRENDE NULLEN VON SPALTE i ABFRAGEN
// ======================================
// zuerst Hauptdiagonalelement abfragen
      if(B.element(i,i) != 0.0)
        for(la=ia;la<=i;la++)
        {
          if(symdef >= 0) { if(B.element(la,i) != 0.0) goto M60;}
          else            { if(B.element(i,la) != 0.0) goto M60;};
        };
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      Y.element(i,k)=0.0;
      continue;
//
M60:  r=R.element(i,k);
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//***      if(symdef >=0)
// Cholesky und Gauss symmetrisch
//***      {
//***        for(l=la;l<i;l++) r=r-B.element(l,i)*Y.element(l,k);
//***      }
//***      else
// Gauss unsymmetrisch -> Untermatrix C zeilenweise
//***      {
//***        for (l=la;l<i;l++) r=r-B.element(i,l)*Y.element(l, k);
//***      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert
// =======================
      if(symdef >=0)
// Cholesky und Gauss symmetrisch
        r-=skalarprodukt(B_.adresse(la,i),1,Y_.adresse(la,k),1,i-la);
// Gauss unsymmetrisch -> Untermatrix C zeilenweise
      else
        r-=skalarprodukt(B_.adresse(i,la),B_.dimi,Y_.adresse(la,k),1,i-la);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ---------------------------------------------------------------------------
// Cholesky
// Hauptdiagonale reziprok gespeichert !!!
      if(symdef > 0) r=r*B.element(i,i);
      Y.element(i,k)=r;
  };
// ===========================================================================
};
// ===========================================================================
//***cout<<"Gestaffelte Rechte Seite= ";  Y.ausgeben(cout);
};// -------------------------------------------------------------------------
//
// Loesungen nach Cholesky oder Gauss ermitteln
//                                       -> S3160(Xmat,Bmat,Ymat,mb,zr,symdef)
// ===========================================================================
void MATRIX::S3160(MATRIZEN &X,MATRIZEN &B,MATRIZEN &Y,int mb,int zr,
                                                                   int symdef)
{
// Loesungen fuer Gleichungsystem nach Cholesky (symmetrisch) oder
// nach Gauss (symmetrisch oder unsymmetrisch) aus---
// Loesungen fuer Gleichungsystem nach Cholesky (symmetrisch) oder
// nach Gauss (symmetrisch oder unsymmetrisch) aus der gestaffelten
// Koeffizientenmatrix und den gestaffelten Rechten Seiten ermitteln

// X(mb,zr) = Matrix fuer voll abgespeicherte Loesungen
// B(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
//            (Obermatrix A bei Gauss unsymmetrisch)
// Y(mb,zr) = Matrix mit gestaffelter Form der Rechten Seiten
//            (Y darf gleich X sein)
// mb       = Zeilenzahl von B,X,Y
// zr       = Spaltenzahl der Loesungen Seiten X,Y
// symdef= 1 : nach Cholesky staffeln
//       = 0 : nach Gauss symmetrisch   staffeln
//       =-1 : nach Gauss unsymmetrisch staffeln
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983
// Compiler: C++ 
// ----------------------------------------------------------------------------
// Algorithmus:
//
// Skalarprodukte
// Gauss symmetrisch: Y(i,k)=Y(i,k)/B(i,i)
// X(i,k)=Y(i,k)-B(i,l)*X(l,k)
// Cholesky und Gauss unsymmetrisch: X(i,k)=X(i,k)/B(i,i)
// ----------------------------------------------------------------------------
// Referenztyp wandeln
MATRIX &X_=this->referenz(X);
MATRIX &B_=this->referenz(B);
MATRIX &Y_=this->referenz(Y);
// --------------------------
// LOKALE VARIABLE
int k,i,le;
double r;
// ---------------------------------------------------------------------------
//*** cout << "Gestaffelte Rechte Seiten Y= "<< Y <<endl;
//*** cout <<"Gestaffelte Koeffizientenmatrix B= "<< B <<endl;
// ===========================================================================
// SCHLEIFE RUECKWAERTS UEBER DIE ZEILEN VON Y,X
// =============================================
for(i=mb-1;i>=0;i--)
{
// ---------------------------------------------------------------------------
// HINTERE NULLEN IN ZEILE i VON B ABFRAGEN
// ========================================
// zuerst Hauptdiagonalelement abfragen
      if(B.element(i,i) != 0.0)
         for(le=mb-1;le>=i;le--) if(B.element(i,le) != 0.0) goto M60;
//
// SPALTE I GESTRICHEN -> Ende Schleife J
      for(k=0;k<zr;k++) X.element(i,k)=0.0;
      continue;
// ---------------------------------------------------------------------------
// Vorbelegung der Loesungen
M60:  if(&X != &Y) for(k=0;k<zr;k++) X.element(i,k)=Y.element(i,k);
//
// Gauss symmetrisch
// Hauptdiagonale reziprok gespeichert !!!
      if(symdef == 0)
        for(k=0;k<zr;k++) X.element(i,k)=X.element(i,k)*B.element(i,i);
// ===========================================================================
// SCHLEIFE UEBER DIE SPALTEN k VON Y
// zr Loesungen simultan ermitteln
// ==================================
  for(k=0;k<zr;k++)
  {
// ---------------------------------------------------------------------------
// SCHLEIFE UEBER DEN SUMMATIONSINDEX l (Skalarprodukt)
// ====================================================
      r=X.element(i,k);
//
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt
// =============
//*** int l;
//***      for(l=le;l>=i+1;l--)
//***      {
//***        r=r-B.element(i,l)*X.element(l,k);
//***      };
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Skalarprodukt optimiert 
// =======================
      r-=skalarprodukt(B_.adresse(i,le),-B_.dimi,X_.adresse(le,k),-1,le-i);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ---------------------------------------------------------------------------
// Cholesky und Gauss unsymmetrisch
// Hauptdiagonale reziprok gespeichert !!!
     if(symdef != 0) r=r*B.element(i,i);
//
     X.element(i,k)=r;
  };
// ---------------------------------------------------------------------------
};
// ===========================================================================
//*** cout<<"Loesungen ";  X_.ausgeben(cout);
};// =========================================================================
//
// Inverse einer   symmetrischen Matrix nach Cholesky oder Gauss
// Inverse einer unsymmetrischen Matrix nach Gauss (ohne Zeilen-/Spalten-
// Vertauschung !!!) bilden
//                                                  -> S3170(Kmat,Bmat,symdef)
// ===========================================================================
double MATRIX::S3170(MATRIZEN& K, MATRIZEN& Z, int mk, int symdef)
{
// Koeffizientenmatrix
// nach Cholesky (nur positiv definite Matrix) oder Gauss invertieren
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// Z(mk,mk) = Matrix fuer die Inverse
//            (Z darf gleich K sein) 
// mk       = Zeilen/Spalten von K, Z
// symdef   =  1 Cholesky
//          =  0 Gauss (symmetrisch)
//          = -1 Gauss (unsymmetrisch)
// Ergebnis: Determinante der Matrix K
// ---------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  13.12.1983 / 07.11.2022 
// Compiler: C++ 
// ---------------------------------------------------------------------------
    // Dimensionen der Matrix Z ermitteln
    int zi, zj, zk, zl;
    Z.dimension_ermitteln(zi, zj, zk, zl);
    // Zeilen/Spalten der Matrix K ermitteln
    int ki, kj, kk, kl;
    mk = K.zeilen_spalten(ki, kj, kk, kl);
    //*** cout << " S3170: mk= " << mk << " ki, kj, kk, kl= " << ki << kj << kk << kl << endl;
    // ------------------------------------------------
    // Matrix-Dimensionen aendern
    K.dimension_aendern(mk, mk);
    Z.dimension_aendern(mk, mk);
    // ------------------------------------------------
    // Zwischenspeicher (permanente Matrix) fuer gestaffelte Form von K aufbauen
    //
    // gestaffelte Obermatrix
    MATRIZEN& Bstaff = *temporaere_matrix_erzeugen(mk, mk, 1, 1, 1);
    // gestaffelte Untermatrix
    MATRIZEN& Cstaff = *temporaere_matrix_erzeugen(mk, mk, 1, 1, 1);
    // -------------------------------------------------------
    // Koeffizientenmatrix staffeln 
    double det = 0.0;
    //
    if (symdef == -1)
    { // Gauss (unsymmetrisch)
         det = S3153(K, Bstaff, Cstaff, mk);
    //*** cout << "S3170: gestaffelte Obermatrix  Bstaff= " << Bstaff << endl;
    //*** cout << "S3170: gestaffelte Untermatrix Cstaff= " << Cstaff << endl;
    }
    else
        if (symdef == 0)
        { // Gauss symmetrisch
             det = S3151(K, Bstaff, mk);
    //*** cout << "S3170: gestaffelte Obermatrix  Bstaff= " << Bstaff << endl;
        }
        else
        { // Cholesky symmetrisch
             det = S3152(K, Bstaff, mk);
    //*** cout << "S3170: gestaffelte Obermatrix  Bstaff= " << Bstaff << endl;
        }
    // ------------------------------------------------
    // Einheitsmatrix Z als Rechte Seiten aufbauen 
    //
    Z = 0.0;
    for (int i = 0; i < mk; i++) Z(i, i) = 1.0;
    //*** cout << "S3170: Einheitsmatrix Z= " << Z << endl;
    //
    // ------------------------------------------------
    // Rechte Seiten staffeln
    if (symdef == -1)
    {
      S3157(Z, Cstaff, Z, mk, mk, symdef);
      //*** cout << "S3170: gestaffelte rechte Seite Zstaff= " << Z << endl;
    }
    else
    {
      S3157(Z, Bstaff, Z, mk, mk, symdef);
      //*** cout << "S3170: gestaffelte rechte Seite Zstaff= " << Z << endl;
    }
// ------------------------------------------------
// Kehrmatrix (Loesungen) Z ermitteln
//
   S3160(Z,Bstaff,Z,mk,mk,symdef);
   //*** cout << "S3170: Inverse Matrix Z= " << Z << endl;
// ------------------------------------------------
// Matrix-Dimensionen zuruecksetzen
K.dimension_aendern(ki, kj, kk, kl);
Z.dimension_aendern(zi, zj, zk, zl);
// ------------------------------------------------
// Zwischenspeicher loeschen
delete &Bstaff;
delete &Cstaff;
// ------------------------------------------------
// ggfl. temporaere Matrizen loeschen
K.temporaere_matrix_loeschen();
Z.temporaere_matrix_loeschen();
// ------------------------------------------------
return det;
};// ========================================================================
//
// Lagerbedingungen in Koeffizientenmatrix einarbeiten
//                                                -> S6115(Kmat,Lmat,Bmat,mk)
// ==========================================================================
void MATRIX::S6115(MATRIZEN &K,MATRIZEN &L,MATRIZEN &B,int mk)
{
// Lagerbedingungen in Koeffizientenmatrix einarbeiten 
// K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
// L(mk)    = Vektor mit Lagerangaben
//            -1 = Starre Lagerung
//             0 = keine  Lagerung
//            >0 = Federkonstante
// B(mk,mk) = Matrix fuer die gelagerte Koeffizientenmatrix
//            (B darf gleich K sein)
// mk       = Zeilen/Spalten von K
//            Zeilen von L
// ----------------------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Hannover
// Version:  14.12.1983
// Compiler: C++ 
// ----------------------------------------------------------------------------
// LOKALE VARIABLE
int i,l;
double lager;
// ---------------------------------------------------------------------------
//***cout<<"Koeffizientenmatrix K= "; K.ausgeben(cout);
//***cout<<"Lagerbedingungen L= ";    L.ausgeben(cout);
// ===========================================================================
// K nach B speichern
if(&K != &B) B=K;
//
// SCHLEIFE UEBER ZEILEN DES LAGERVEKTORS
for(i=0;i<mk;i++)
{ lager=L.element(i);
 
  if(lager > 0.0)
  {
    B.element(i,i)=K.element(i,i)+lager; 
  }
  else
    if(lager < 0.0)
    {
      for(l=0;l<mk;l++)
      {
        B.element(i,l)=0.0;
        B.element(l,i)=0.0;      
      };
    };
};
//***cout<<"Gelagerte Koeffizientenmatrix B= "; B.ausgeben(cout);
};
// ======================================================================================
